Pinocchio VM
============

Pinocchio is a virtual machine which natively implements a variation of the
Smalltalk programming language in which the interpreter itself is
meta-circularly present. This meta circular interpreter allows developers or
researchers to experiment with programming languages variations and debuggers
at low cost. Instead of changing the code of the VM written in C, the
meta-circular interpreter can be sub-classed and manipulated as any other OO
entity — code running at the application-level and the interpreter-level belong
then to the same object space. This approach overcomes limitations of other
reflection-based systems for which programming language variations are
inherently restricted by the features of interpreter itself.

A full package including the sources and the current development environment
can be found at http://scg.unibe.ch/download/pinocchio/.

For code see http://github.com/pinocchio/p/ and for bugs see
http://code.google.com/p/pinocchiovm/issues/list.

A further introduction is given by our paper 
[Bringing Reflection to Life with First-Class Interpreters](http://dx.doi.org/10.1145/1869459.1869522)
published at Onward! 2010, Reno/Tahoe, Nevada, USA.


Demos
=====

To run the demos first compile the sources. This release of Pinocchio requires
the Boehm garbage collector a helper script for OSX is provided at
`meta/mac_install_boehm`. Once installed proceed with the following
instructions:

    > unzip pinocchio.zip
    > cd pinocchio/src
    > make clean
    > make -j3
    > ./pinocchio

The C sources are all included in `src/` whereas the Smalltalk sources are
included in the Pharo image located at `pharo/Pharo-1.1.image`. VMs for Linux,
Mac and Windows can be found in the same directory.

To regenerate the sources launch the exporter from within Pharo by evaluating 
`P exportAll`.


Debugger
--------

    > Debugger interpret: [ Person new ]
    BlockClosure>>#value (0)
    pidb> 
                  Person class>>#new (0)
    pidb> 
                    ObjectLayout>>#instantiate: (1)
    pidb> 
                      ObjectLayout>>#basicInstantiate: (1)
    pidb> 
                      ObjectLayout>>#initialize: (1)
    pidb> 
                    Person>>#initialize (0)
    pidb> 
    Person
            name: Nil
    > 

Object Flow Interpreter
-----------------------

    > alias := AliasInterpreter interpret: [
    |   person := Person new.
    |   person name: 'John'.
    |   person name: 'Doe'.
    |   person
    | ]
    ReturnAlias (Shell>>DoIt)
            origin: ReturnAlias (Behavior>>new)
            context: MethodContext
            value: Person
    > alias value
    Person
            name: FieldWriteAlias (Person>>name:)
    > alias value name
    FieldWriteAlias (Person>>name:)
            origin: ParameterAlias (Shell>>DoIt)
            context: MethodContext
            value: 'Doe'
            predecessor: FieldWriteAlias (Person>>name:)
    > alias value name predecessor
    FieldWriteAlias (Person>>name:)
            origin: ParameterAlias (Shell>>DoIt)
            context: MethodContext
            value: 'John'
            predecessor: AllocationAlias (ObjectLayout>>basicInstantiate:)
    > alias value name predecessor predecessor
    AllocationAlias (ObjectLayout>>basicInstantiate:)
            origin: Nil
            context: NativeMethodContext
            value: Nil

